/*
 * $Id: //trid/release/pam_otp_auth-3.2.2/otp_pam.c#1 $
 *
 * PAM authentication management functions for pam_otp_auth.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 *
 *  For alternative licensing, contact licensing@tri-dsystems.com.
 *
 * Copyright 2002-2005  Google, Inc.
 * Copyright 2005,2006 TRI-D Systems, Inc.
 *
 */

#include "ident.h"
RCSID("$Id: //trid/release/pam_otp_auth-3.2.2/otp_pam.c#1 $")

#include <security/pam_appl.h>
#include <security/pam_modules.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>

#include "extern.h"
#include "otp.h"	/* OTP_MAX_CHALLENGE_LEN */

/* Authenticate a user via otp token */
int
pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
  int	rc;
  char	*service;
#if 0
  void	*service;	/* void * due to gcc aliasing deficiency */
#endif
  char	*username;
  char	*prompt = NULL, *passcode = NULL;
  int	pass_num;
  char	challenge[OTP_MAX_CHALLENGE_LEN + 1];	/* +1 for '\0' terminator */

  otp_option_t opt;

  *challenge = '\0';

  /* get service name */
  (void) pam_get_item(pamh, PAM_SERVICE, (PAM_GET_CONST void **) &service);
  if (!service)
    service = "unknown";

  if ((rc = otp_parse_opts(argc, argv, service, &opt)) != PAM_SUCCESS)
    goto auth_done;

  /* Get username */
  if ((rc = pam_get_user(pamh, (PAM_GET_CONST char **) &username,
                         "login: ")) != PAM_SUCCESS) {
    if (opt.debug)
      syslog(LOG_DEBUG, "pam_otp_auth(%s): %s: pam_get_user: %s",
             service, __func__, pam_strerror(pamh, rc));
    goto auth_done;
  }

  /*
   * Get passcode.  We use a counter (pass_num) since in fast_sync mode,
   * the user might explicitly request a challenge; i.e., getting a response
   * from the user does not necessarily mean it's the passcode.
   */
  if (opt.useauthtok) {
    /* undocumented use_first_pass functionality, useful for testing */
    if ((rc = pam_get_item(pamh, PAM_AUTHTOK,
                           (PAM_GET_CONST void **) &passcode)) != PAM_SUCCESS) {
      if (opt.debug)
        syslog(LOG_DEBUG, "pam_otp_auth(%s): %s: pam_get_item: %s",
               service, __func__, pam_strerror(pamh, rc));
      goto auth_done;
    }
    pass_num = 2;	/* avoid while() loop */
  } else {
    pass_num = 0;
  }
  while (pass_num <= 1) {
    rc = PAM_SERVICE_ERR;
    if (opt.fast_sync) {
      prompt = malloc(strlen(opt.fast_prompt) + 1);
      if (!prompt) {
        syslog(LOG_CRIT, "pam_otp_auth(%s): %s: out of memory!",
               service, __func__);
        goto auth_done;
      }
      (void) strcpy(prompt, opt.fast_prompt);
    } else {
      /* get a random challenge */
      if (otp_async_challenge(challenge, opt.chal_len) == -1) {
        syslog(LOG_ERR,
               "pam_otp_auth(%s): %s: failed to obtain random challenge",
               service, __func__);
        goto auth_done;
      }

      /*
       * A little more than needed since 2 bytes of the chal_prompt
       * are replaced by 2 bytes of the challenge.
       */
      prompt = malloc(strlen(opt.chal_prompt) + opt.chal_len + 1);
      if (!prompt) {
        syslog(LOG_CRIT, "pam_otp_auth(%s): %s: out of memory!",
               service, __func__);
        goto auth_done;
      }
      (void) sprintf(prompt, opt.chal_prompt, challenge);

      pass_num++; /* in C/R we always get the passcode, so exit loop */
    }

    /* We must always set ECHO_OFF because of soft PIN. */
    rc = otp_interact(pamh, prompt, PAM_PROMPT_ECHO_OFF, &passcode);
    free(prompt);
    if (rc) {
      if (opt.debug)
        syslog(LOG_DEBUG, "pam_otp_auth(%s): %s: otp_interact() failed: %s",
               service, __func__, pam_strerror(pamh, rc));
      goto auth_done;
    }

    if (opt.fast_sync) {
      /* prompt with challenge if requested */
      if (opt.allow_async && !strcmp(passcode, opt.chal_req)) {
        opt.fast_sync = 0;	/* force challenge generation */
        free(passcode); passcode = NULL;
      } else {
        pass_num++;		/* we got the sync response, exit loop */
      }
    }

    pass_num++;
  } /* while (pass_num <= 1) */

  /* do it */
  rc = otp_pw_valid(username, passcode, challenge, &opt);

  /* Must exit here after calling parse_opts(). */
auth_done:
#ifdef PAM_AUTH_ERROR
  /*
   * Check the PAM_AUTH_ERROR flag.  We don't do this earlier because
   * we want to resync the card even if authentication will ultimately
   * fail, to prevent a replay attack.  This flag has been documented
   * on Solaris for a long time, but [apparently] never implemented.
   */
  if (flags & PAM_AUTH_ERROR) {
    if (opt.debug)
      syslog(LOG_DEBUG, "pam_otp_auth(%s): %s: PAM_AUTH_ERROR set",
             service, __func__);
    rc = PAM_AUTH_ERR;
  }
#endif /* PAM_AUTH_ERROR */

  if (opt.debug)
    syslog(LOG_DEBUG, "pam_otp_auth(%s): %s: %s%s", service, __func__,
            rc == PAM_SUCCESS ? "exit success" : "exit failure ",
            rc == PAM_SUCCESS ? "" : pam_strerror(pamh, rc));

  otp_free_opts(&opt);
  if (passcode)
    free(passcode);

  /*
   * Modules are supposed to set PAM_AUTHTOK if not set, but no other
   * module should be interested in the OTP, so we don't do that.
   */
  return rc;
}


/* Called after a successful authentication to set user credentials. */
int
pam_sm_setcred(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
  return PAM_SUCCESS;
}
